/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    Navigation::ArmarXObjects::Navigator
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "Navigator.h"
#include "ArmarXCore/core/services/tasks/TaskUtil.h"

// Navigation
#include "Navigation/libraries/core/types.h"
#include "Navigation/libraries/util/util.h"
#include <Navigation/libraries/factories/NavigationStackFactory.h>
#include <Navigation/libraries/server/Navigator.h>

namespace armarx::nav::components
{

    Navigator::~Navigator()
    {
    }

    void Navigator::onInitComponent()
    {
    }

    void Navigator::onConnectComponent()
    {
    }

    void Navigator::onDisconnectComponent()
    {
    }

    void Navigator::onExitComponent()
    {
    }
    
    void Navigator::updateContext() 
    {
        ctx.staticScene = staticScene();
        ctx.robot = getRobot();
    }

    std::string Navigator::getDefaultName() const
    {
        return "Navigator";
    }

    void Navigator::moveTo(const Eigen::Matrix4f& pose,
                           const aron::data::AronDictPtr& stackConfig,
                           const std::string& navigationMode,
                           const Ice::Current&)
    {
        updateContext();

        server::NavigationStack stack = fac::NavigationStackFactory::create(stackConfig, ctx);
        server::Navigator navigator{stack, ctx, executor};

        SimplePeriodicTask<> task(
            [](){navigator.updateContext()}
        );
        task.start()
        

        navigator.moveTo(core::Pose(pose), core::NavigationFramesMap.from_name(navigationMode));
    }

    void Navigator::moveTowards(const Eigen::Matrix4f& direction,
                                const aron::data::AronDictPtr& stackConfig,
                                const std::string& navigationMode,
                                const Ice::Current&)
    {
        server::NavigationStack stack = fac::NavigationStackFactory::create(stackConfig, ctx);
        server::Navigator navigator{stack, ctx};

        // navigator.moveTowards(direction); TODO
    }

    armarx::PropertyDefinitionsPtr Navigator::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr def =
            new ComponentPropertyDefinitions(getConfigIdentifier());

        // Publish to a topic (passing the TopicListenerPrx).
        // def->topic(myTopicListener);

        // Subscribe to a topic (passing the topic name).
        // def->topic<PlatformUnitListener>("MyTopic");

        // Use (and depend on) another component (passing the ComponentInterfacePrx).
        def->component(platformUnit);

        // Add a required property.
        // def->required(properties.boxLayerName, "p.box.LayerName", "Name of the box layer in ArViz.");

        // Add an optionalproperty.
        // def->optional(properties.numBoxes, "p.box.Number", "Number of boxes to draw in ArViz.");

        return def;
    }

    core::StaticScenePtr Navigator::staticScene()
    {
        core::StaticScenePtr scene(new core::StaticScene);

        objpose::ObjectPoseSeq objectPoses = ObjectPoseClientPluginUser::getObjectPoses();

        scene->objects = util::asSceneObjects(objectPoses);

        return scene;
    }
    
    VirtualRobot::RobotPtr Navigator::getRobot() 
    {

        auto robot = RemoteRobot::createLocalCloneFromFile(
            getRobotStateComponent(), VirtualRobot::RobotIO::RobotDescription::eFull);
        // auto robot = RemoteRobot::createLocalClone(getRobotStateComponent());
        ARMARX_CHECK_NOT_NULL(robot);

        // ARMARX_INFO << "Resetting robot";
        // robot = VirtualRobot::RobotIO::loadRobot(robot->getFilename());

        RemoteRobot::synchronizeLocalClone(robot, getRobotStateComponent());

        return robot;
    }

} // namespace armarx::nav::components
