#include "Navigator.h"

#include <VirtualRobot/Robot.h>

#include "ArmarXCore/core/exceptions/local/ExpressionException.h"

#include "Navigation/libraries/core/types.h"

namespace armarx::nav::server
{

    // Stack result
    core::TrajectoryPtr StackResult::trajectory() const
    {
        if (localTrajectory)
        {
            return localTrajectory;
        }

        ARMARX_CHECK_NOT_NULL(globalTrajectory);
        return globalTrajectory;
    }

    core::Twist StackResult::velocity() const
    {
        if (safeVelocity.has_value())
        {
            return safeVelocity.value();
        }

        ARMARX_CHECK(controlVelocity.has_value());
        return *controlVelocity;
    }

    // Navigator

    Navigator::Navigator(const server::NavigationStack& navigationStack,
                         const core::Scene& scene,
                         ExecutorInterface& executor) :
        stack(navigationStack), scene(scene), executor{executor}
    {
        setTag("Navigator");
    }

    void Navigator::moveTo(const core::Pose& goal, const core::NavigationFrames& navigationFrame)
    {
        core::Pose globalPose;
        switch (navigationFrame)
        {
            case core::NavigationFrames::Absolute:
                globalPose = goal;
                break;
            case core::NavigationFrames::Relative:
                globalPose = core::Pose(scene.robot->getGlobalPose()) * goal;
                break;
        }

        moveToAbsolute(globalPose);
    }

    void Navigator::moveToAbsolute(const core::Pose& goal)
    {
        ARMARX_INFO << "Planning global trajectory";
        ARMARX_CHECK_NOT_NULL(stack.globalPlanner);

        StackResult res;
        res.globalTrajectory = stack.globalPlanner->plan(goal);

        if (stack.localPlanner)
        {
            ARMARX_INFO << "Planning local trajectory";

            // TODO: start the local planner thread
            res.localTrajectory = stack.localPlanner->plan(res.globalTrajectory);
        }

        ARMARX_CHECK_NOT_NULL(stack.trajectoryControl);
        while (true)
        {
            ARMARX_INFO << "Trajectory control";

            res.controlVelocity = stack.trajectoryControl->control(res.trajectory());

            if (stack.safetyControl)
            {
                ARMARX_INFO << "Safety control";
                res.safeVelocity = stack.safetyControl->control(res.controlVelocity.value());
            }

            executor.move(res.velocity());
        }
    }
} // namespace armarx::nav::server
